# Formal Verification of Storm - Supplementary Material

## INTRODUCTION

This repository contains the supplementary material accompanying the paper named:
**Formally Verifying the Absence of Overloading in Apache Storm Topologies**


## CONTENT


### SOFTWARE

- ```storm_v1_nusmv.c```:  interactive program in C language to generate NuSMV models.  It takes in input the parameters (ES, EB, number of threads) of the nodes and the topology information, and produces the corresponding runnable NuSMV model.
   - WARNING: the software isn't parameterized in the number of bins (yet); it produces models with 3 bins
   - WARNING: the produced model is named *Spec.smv*; remember to rename it before producing another model in order to avoid unwanted overwriting

- ```expFM2024.bat```:  script (batch) file for Windows to verify a model with NuSMV in batch mode.
   - USAGE:  expFM2024.bat  \<modelname\>.smv
   - WARNING:  please update the pathname of your NuSMV executable in the script, before the first use

- ```testFM24bat.src```:  NuSMV script sourced by expFM2024.bat to execute a model and produce the verification statistics.

- ```testFM24int.src```:  NuSMV script; it has the same behaviour as testFM24bat.src and can be used with NuSMV in interactive mode to execute a model and produce the verification statistics.


### MODELS

- ```MODELS.zip```:  archive file containing 131 models, most produced with storm_v1_nusmc.c apart for those manually produced for the 4-bins examples.
   - NAME PATTERN 1:  \<#bins\>b_ES\<value\>_EB\<value\>_t\<#threads per bolt\>.smv
      - used when all spouts have the same periods, all bolts include the same number of threads, and all threads of all bolts have the same processing time
   - NAME PATTERN 2:  \<#bins\>b_ES\<valueS1\>-...-\<valueSi\>_EB\<value\>_t\<#threadsB1\>...\<#threadsBj\>
      - used when spouts have different periods and bolts have different number of threads


### RESULTS

- ```resultsFM24.xlsx```:  spreadsheet with the verification and performance results of each model in MODELS.zip.  In addition to the indices analysed in the paper, the following indices are extracted (with excerpts from the CUDD Decision Diagram documentation):
   - **Peak number of nodes**:  "This number is the number of nodes that the manager has allocated. This is not the largest size of the BDDs, because the manager will normally have some dead nodes and some nodes on the free list."
   - **Peak number of live nodes**: "This is the largest number of live nodes that the manager has held since its creation."
   - **Number of BDD and ADD nodes**: "How many nodes are currently in the unique table, either alive or dead."
   - **Number of dead BDD and ADD nodes**: "Subtract these numbers from those above to get the number of live nodes."
   - **Total number of nodes allocated**: "This is the total number of nodes that were requested and obtained from the free list. It never decreases, and is not an indication of memory occupation after the first garbage collection. Rather, it is a measure of the package activity."
   - **Total number of nodes reclaimed**: "These are the nodes that were resuscitated from the dead.  If they are many more than the allocated nodes, and the total number of slots is low relative to the number of nodes, then one may want to increase the limit for fast unique table growth. In this case, the number is 0 because of delayed dereferencing."
   - nodes = dead + live


## DISCLAIMER

This software is free and, as such, it comes with no warranty. 
By using this software, you are accepting the terms stated here. 
Because the program is licensed free of charge, there is no warranty for the program, to the extent permitted by applicable law. 
The entire risk as to the quality and performance of the program is with you. 
Should the program prove defective, you assume the cost of all necessary servicing, repair or correction directly or indirectly derived from its use.

