

/*
compile with:

gcc -o storm storm.c  

*/

#include <string.h>
#include<stdio.h>
#include<stdlib.h>
#include<stdarg.h>
#include<memory.h>
#include<setjmp.h>
#include <ctype.h>
#include <getopt.h>
#include <stdbool.h>


#define NUMBINS   3
#define MAXSPOUTS 10
#define MAXBOLTS  10
#define MAXTIMER  20
#define MAXTHREADS 11
#define MAXQUEUE  5
#define MAXLEAP   5



int main(){
  
   
   int Nbolts; int Nspouts;

   printf("\nAPACHE STORM VERIFICATION WITH 3 BINS \n\n");
   
   printf("\nSpecify the number of spouts (max %d): ",MAXSPOUTS);
   scanf("%d", &Nspouts);
   if(Nspouts>MAXSPOUTS){
        printf("\nToo large value!\n\n"); exit(0);
   }
    
   printf("\nSpecify the number of bolts (max %d): ",MAXBOLTS);
   scanf("%d", &Nbolts);
   if(Nbolts>MAXBOLTS){
        printf("\nToo large value!\n\n"); exit(0);
   }
   
   int spout_timers[MAXSPOUTS+1];
   int bolt_queues[MAXBOLTS+1];
   int bolt_threads[MAXBOLTS+1];
   int bolt_timers[MAXBOLTS+1];
   int emit_sb[MAXSPOUTS+1][MAXBOLTS+1];
   int emit_bb[MAXBOLTS+1];
 
 
   int I; int J; int K; int tleap;
   bool ok = false;

   for(I=1;I<=Nbolts;I++){
     printf("\nSpecify the max allowed queue of bolt %d (max %d): ",I,MAXQUEUE);
     scanf("%d", &bolt_queues[I]);
     if(bolt_queues[I]>MAXQUEUE){
        printf("\nToo large value!\n\n"); exit(0);
        }
   }

   for(I=1;I<=Nbolts;I++){
     printf("\nSpecify the  number of threads of bolt %d (max %d): ",I, MAXTHREADS-1);
     scanf("%d", &bolt_threads[I]);
     if(bolt_threads[I]>=MAXTHREADS){
        printf("\nToo large value!\n\n"); exit(0);
        }
   }

   for(I=1;I<=Nbolts;I++){
     do {
        ok = true;
        printf("\nSpecify the processing time of threads of bolt %d (max %d): ",I, MAXTIMER);
        scanf("%d", &bolt_timers[I]);
        if(bolt_timers[I]>MAXTIMER){
           printf("\nToo large value!\n\n"); ok=false; }
        if(bolt_timers[I] < NUMBINS || (bolt_timers[I] % NUMBINS) != 0) {
           printf("\nTime must be a multiple of %d!\n\n",NUMBINS); ok=false; }
     }while(!ok);
   }
   tleap = bolt_timers[1] / NUMBINS;
   printf("Factor of time acceleration: %d\n",tleap);

   for(I=1;I<=Nspouts;I++){
     do {
        ok = true;
        printf("\nSpecify the emit interval of spout %d (max %d): ",I,MAXTIMER);
        scanf("%d", &spout_timers[I]);
        if(spout_timers[I]>MAXTIMER){
           printf("\nToo large value!\n\n"); ok=false; }
        if(spout_timers[I] < tleap || (spout_timers[I] % tleap) != 0) {
           printf("\nTime must be a multiple of %d!\n\n",tleap); ok=false; }
     }while(!ok);
   }

 
 for(I=1;I<=Nspouts;I++){
      for(J=1;J<=Nbolts;J++){
         char c; char c1;
         printf("\nSpecify whether spout n.%d can emit to bolt %d (y/n, def.n): ",I,J);
         scanf("%c%c1", &c,&c1);
         if(c1=='y' || c1=='Y'){emit_sb[I][J]=1;}else{emit_sb[I][J]=0;}
         //printf("\n%c+%c\n",c,c1);         
      }   
   }
   

  printf("\n\nNow insert the emit relations among bolts;\n");

  printf("when specifying the emission bolts be aware that cycles may produce misbehavior.\n");
  for(I=1;I<=Nbolts;I++){
         printf("\nSpecify the emission bolt of bolt %d (0 if it is terminal): ",I);
         scanf("%d", &emit_bb[I]);
         if(emit_bb[I]>Nbolts){printf("\nBolt %d does not exist!\n",emit_bb[I]);exit(0);}
         if(emit_bb[I]==I){printf("\nA bolt cannot emit to itself!\n");exit(0);}
    }
   
 printf("\nElaborating the specification file...\n\n");
 
 
 
 char *backupFile = "Spec.smv";
  FILE *fq;
  fq=fopen(backupFile, "w");
   if (!fq) {
       fprintf(stderr, "Error! Cannot open file %s\n", backupFile);
			 exit(1);
		       }
   fprintf(fq,"\n\nMODULE main\n\nDEFINE");
 
   fprintf(fq,"\n--factor of time acceleration \n");   
   fprintf(fq,"leap:= %d;",tleap);
 
   fprintf(fq,"\n--actual number of threads in the various bolts");   
   for(I=1;I<=Nbolts;I++){
     fprintf(fq,"\nthreads_in_B%d:= %d;", I,bolt_threads[I]);
    }
    
   fprintf(fq,"\n--maximum queues in the various bolts");   
   for(I=1;I<=Nbolts;I++){
     fprintf(fq,"\nqueue_in_B%d:= %d;", I,bolt_queues[I]);
    }
 
   fprintf(fq,"\n--emission time in the various spouts");
   for(I=1;I<=Nspouts;I++){
      fprintf(fq,"\nemission_in_S%d:= %d;", I,spout_timers[I] );
   }

   fprintf(fq,"\n--emission time in the various bolts");
   for(I=1;I<=Nbolts;I++){
      fprintf(fq,"\nemission_in_B%d:= %d;", I,bolt_timers[I] );
   }

   fprintf(fq, "\n\nerror:= ");
   for(I=1;I<=Nbolts-1;I++){
     fprintf(fq,"B%d.Q>queue_in_B%d | ",I, I);
   }
   fprintf(fq,"B%d.Q>queue_in_B%d; \n",Nbolts,Nbolts);    
 
    //fprintf(fq,"\nSPEC \n  AG !error;\n");
   fprintf(fq,"\n\nLTLSPEC \n  G !error;\n");
 
 
   fprintf(fq,"\nVAR\n--phi is the step\n phi:  0..5;\n");
   fprintf(fq,"-- phi=0 this is the spout emission step (if enough time has passed);\n");
   fprintf(fq,"-- phi=1 bolt queues are updated after spout emission;\n");
   fprintf(fq,"-- phi=2 (was 3) bolt i threads that finished execution (enumerated in Bi.100) becomes idle and emit tuples; \n");
   fprintf(fq,"-- phi=3 (was 4) bolt queues are updated after bolt emission;\n");
   fprintf(fq,"-- phi=4 (was 2) idle threads of  bolt i (in Bi.I) takes care of as much as possible tuples (in Bi.Q) and go to the exec0 state (in Bi.0); \n");
   fprintf(fq,"-- phi=5 time elapse (one execution phase in each bolt i: Bi.0-->Bi.33-->Bi.67-->Bi.100)\n"); 
     
               
   fprintf(fq,"\n--spout modules\n");
   fprintf(fq,"--the third parameter is the emission interval");

   
   
   for(I=1;I<=Nspouts;I++){
     fprintf(fq,"\nS%d: spout(phi,leap,emission_in_S%d",I,I); 
     for(J=1;J<=Nbolts;J++){  
        if(emit_sb[I][J]==1)fprintf(fq,",TRUE"); else fprintf(fq,",FALSE");
     }
     fprintf(fq,"); --spout %d emit to bolts", I);
     for(J=1;J<=Nbolts;J++){  
        if(emit_sb[I][J]==1)fprintf(fq," %d",J); 
     } 
   }
   
   fprintf(fq,"\n--bolt modules: last parameter is the allowed length of the queue");
   for(I=1;I<=Nbolts;I++){
   //BI: bolt(I,phi,S1.EI,S2.EI,..,B1.EI,B2.EI,B3.EI,...,...FALSE/TRUE...,nthreads_in_BI);   -- BI emits to ...BJ...
     fprintf(fq,"\nB%d: bolt(%d,phi",I,I);
     for(J=1;J<=Nspouts;J++){fprintf(fq,",S%d.E%d",J,I);}
     for(J=1;J<=Nbolts;J++){fprintf(fq,",B%d.E%d",J,I);}
     for(J=1;J<=Nbolts;J++){
        if(emit_bb[I]==J)fprintf(fq,",TRUE"); else fprintf(fq,",FALSE");
     }   
     fprintf(fq,",emission_in_B%d,threads_in_B%d,queue_in_B%d);",I,I,I);
     int K=0; 
     for(J=1;J<=Nbolts;J++){
        if(emit_bb[I]==J) K=J; 
     } 
     if(K==0)fprintf(fq," --bolt B%d does not emit ",I);  
        else fprintf(fq," --bolt B%d emit to bolt %d ",I, K);
     
   }


  fprintf(fq,"\n\nASSIGN \n init(phi):= 0;\n next(phi):=\n  case\n   phi<5 : phi+1;\n   phi=5 : 0;\n   TRUE  : phi;\n  esac;\n\n");
  
 
  
  fprintf(fq,"\n\nMODULE spout(phi,TL,ES");
  for(I=1;I<=Nbolts;I++){
     fprintf(fq,",b%d",I);
  }    
  fprintf(fq,")\n VAR\n --TS is the Spout clock\n TS: 0..%d;",MAXTIMER);
  fprintf(fq,"\n -- emit indicates to which bolt bi the spout nondeterministically currently emit (0 for no emission)");
  fprintf(fq,"\n -- emission will be effective only in case the channel bi is open");
  fprintf(fq,"\n emit: 0..%d;\n",Nbolts);
  fprintf(fq,"\n-- next variables indicate whether the emission is effective");
  for(I=1;I<=Nbolts;I++){
    fprintf(fq,"\n E%d: {0,1};",I);
   }
   
   
   
  fprintf(fq,"\n\nASSIGN\n init(TS):= 0;\n next(TS):=\n  case\n   phi=0 & TS>=ES   : 0;\n   phi=5 & TS<ES    :TS+TL;\n   TRUE             :TS;\n  esac;");
  
  
  for(I=1;I<=Nbolts;I++){
   fprintf(fq,"\n\nASSIGN");
   fprintf(fq,"\n init(E%d):= 0;",I);
   fprintf(fq,"\n next(E%d):=",I);
   fprintf(fq,"\n  case");
   fprintf(fq,"\n   phi=0 & TS>=ES & b%d & emit=%d   : 1;",I,I);
   fprintf(fq,"\n  TRUE                           : 0;");
   fprintf(fq,"\n     esac;");  
  }
  

  fprintf(fq,"\n\n\n--------------------------");
  fprintf(fq,"\n\nMODULE bolt(id,phi");
  for(I=1;I<=Nspouts;I++){
    fprintf(fq,",rs%d",I);
  }
  for(I=1;I<=Nbolts;I++){
    fprintf(fq,",rb%d",I);
  }
  for(I=1;I<=Nbolts;I++){
    fprintf(fq,",eb%d",I);
  }

  fprintf(fq,",tb,Nthreads,error)");
  fprintf(fq,"\n VAR\n");
  
  fprintf(fq,"\n Q:    0..%d;  --queue for this bolt\n TI:   0..%d;  --idle threads\n T0:   0..%d;  --starting exec threads\n T33:  0..%d;  --one third exec threads\n T67:  0..%d;  --two third exec threads\n T100: 0..%d;  --finished exec threads",MAXTHREADS,MAXTHREADS,MAXTHREADS,MAXTHREADS,MAXTHREADS,MAXTHREADS);
 
  for(I=1;I<=Nbolts;I++){
     fprintf(fq,"\n E%d:    0..%d;   --(possibly) emitted tuples to bolt B%d",I,MAXTHREADS,I);
  }   
  
  
  
  fprintf(fq,"\n\nASSIGN");
  fprintf(fq,"\n  init(Q):= 0;\n  next(Q):=\n   case\n    phi=1 & Q");
  for(I=1;I<=Nspouts;I++)fprintf(fq,"+rs%d",I);
  fprintf(fq,"<=error           : Q");
  for(I=1;I<=Nspouts;I++)fprintf(fq,"+rs%d",I);
  fprintf(fq,";");
  fprintf(fq,"\n    phi=1 & Q");
  for(I=1;I<=Nspouts;I++)fprintf(fq,"+rs%d",I);
  fprintf(fq,">error           : error+1;");
  fprintf(fq,"\n    Q>0  & TI>=Q & phi=4              : 0;"); 
  fprintf(fq,"\n    Q>0  & TI<Q  & phi=4              : Q+ (-1 * TI);");
  fprintf(fq,"\n    phi=3 & Q");
  for(I=1;I<=Nbolts;I++)fprintf(fq,"+rb%d",I);
  fprintf(fq,"<=error           : Q");
  for(I=1;I<=Nbolts;I++)fprintf(fq,"+rb%d",I);
  fprintf(fq,";");
  fprintf(fq,"\n    phi=3 & Q");
  for(I=1;I<=Nbolts;I++)fprintf(fq,"+rb%d",I);
  fprintf(fq,">error           : error+1;"); 
  fprintf(fq,"\n    TRUE                              : Q;\n   esac;\n");
   

  fprintf(fq,"\n\nASSIGN\n init(T100):= 0;\n next(T100):=\n  case\n    phi=2                     : 0;\n    phi=5 	: T67;\n   TRUE                               : T100;\n  esac;\n");

  fprintf(fq,"\n\nASSIGN\n init(T67):= 0;\n next(T67):=\n  case\n    phi=5                      : T33;\n    TRUE                      : T67;\n    esac;\n");
    
  fprintf(fq,"\n\nASSIGN\n init(T33):= 0;\n next(T33):=\n  case\n    phi=5                      : T0;\n    TRUE                      : T33;\n    esac;\n");
 
 
  
  fprintf(fq,"\n\nASSIGN\n init(T0):= 0;\n next(T0):=\n  case");
  fprintf(fq,"\n-- the guards of case 1,2 always evaluate to true for phi=4");
  fprintf(fq,"\n    TI<Q & phi=4  & T0+TI<=%d     : T0+TI;",MAXTHREADS);
  fprintf(fq,"\n    TI>=Q & phi=4 & T0+Q<=%d     : T0+Q;",MAXTHREADS);
  fprintf(fq,"\n    phi=5                      : 0;");
  fprintf(fq,"\n    TRUE                         : T0;\n  esac;\n");
  
 
  fprintf(fq,"\n\nASSIGN\n init(TI):= Nthreads;\n  next(TI):=\n   case\n    Q>0 & TI>=Q & phi=4              : TI + (-1 * Q);\n    Q>0 & TI<Q  & phi=4              : 0;");
  fprintf(fq,"\n-- the guard below always evaluates to true for phi=2");
  fprintf(fq,"\n    phi=2 & TI+T100<=%d            : TI+T100;\n    TRUE                             : TI;\n  esac;\n",MAXTHREADS);
  
 
 for(I=1;I<=Nbolts;I++){
  fprintf(fq,"\n\nASSIGN\n init(E%d):= 0;\n next(E%d):=\n  case",I,I);
  fprintf(fq,"\n       phi=2 & eb%d          : T100;",I);
  fprintf(fq,"\n       TRUE                  : 0;"); 
  fprintf(fq,"\n   esac;"); 
 }
 
  
  
  fprintf(fq,"\n\n");
  
 




  printf("\n..output written on file Spec.smv; run\n  $ NuSMV Spec.smv\nto test it.\n\n");
  
  fclose(fq);

  return 0;
}
